#!/usr/bin/env python
# coding: utf-8

#########################################################################
#########################################################################

"""
   File Name: _cext.pyx
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Wed Nov  5 16:32:11 2014 CST
"""
DESCRIPTION = """
"""

# import both numpy and the Cython declarations for numpy
import numpy as np
cimport numpy as cnp

# use the Numpy-C-API from Cython
cnp.import_array()

# cdefine the signature of our c function
cdef extern from "cext.h":
    void sumidxtab_core_cfunc(const cnp.float32_t * D, const cnp.uint8_t * blk,
        int nsq, int ksub, int cur_num, cnp.float32_t * out)
    void hamming_core_cfunc(cnp.uint8_t * db, cnp.uint8_t * qry, int dim, int num,
                            cnp.uint16_t * dist)
    void knn_count_core_cfunc(cnp.uint16_t * D, int numD, int maxD,
                              int topk, cnp.int32_t * out);
    void fast_euclidean_core_cfunc(const cnp.float32_t * feat, const cnp.float32_t * query,
                                   const cnp.float32_t * featl2norm, int dim, int num,
                                   cnp.float32_t * dist);

# create the wrapper code, with numpy type annotations
def sumidxtab_core(cnp.ndarray[cnp.float32_t, ndim=2, mode="c"] D not None,
                   cnp.ndarray[cnp.uint8_t, ndim=2, mode="c"] blk not None):
    out = np.zeros(blk.shape[0], dtype=D.dtype)
    sumidxtab_core_cfunc(<cnp.float32_t*> cnp.PyArray_DATA(D),
                         <cnp.uint8_t*> cnp.PyArray_DATA(blk),
                         D.shape[0], D.shape[1], blk.shape[0],
                         <cnp.float32_t*> cnp.PyArray_DATA(out))
    return out


def hamming(cnp.ndarray[cnp.uint8_t, ndim=2, mode="c"] Q not None,
            cnp.ndarray[cnp.uint8_t, ndim=2, mode="c"] D not None):
    """
    Calculate Hamming Distance
    """
    nq, dq = Q.shape[0], Q.shape[1]
    nd, dd = D.shape[0], D.shape[1]
    out = np.zeros((nq, nd), dtype=np.uint16)
    dist = np.zeros(nd, dtype=np.uint16)
    for i in xrange(nq):
        hamming_core_cfunc(<cnp.uint8_t*> cnp.PyArray_DATA(Q[i, :]),
                           <cnp.uint8_t*> cnp.PyArray_DATA(D),
                           dq, nd,
                           <cnp.uint16_t*> cnp.PyArray_DATA(out[i, :]))
    return out

def knn_count(cnp.ndarray[cnp.uint16_t, ndim=1, mode="c"] D not None,
              int maxD, int topk):
    """
    Return topk by counting sort
    """
    out = np.zeros(topk, dtype=np.int32)
    numD = D.shape[0]
    knn_count_core_cfunc(<cnp.uint16_t*> cnp.PyArray_DATA(D),
                         D.shape[0], maxD, topk,
                         <cnp.int32_t*> cnp.PyArray_DATA(out))
    return out


def fast_euclidean(feat, query, featl2norm):
    """ Euclidean distance.
    Args:
        feat:       N x D feature matrix
        query:      1 x D feature vector
        featl2norm: 1 x N vector

    Returns:
        dist:       1 x N vector
    """
    N = feat.shape[0]
    D = feat.shape[1]
    dist = np.empty((1, N), dtype=np.float32)
    # dist = featl2norm.astype(np.float32)
    fast_euclidean_core_cfunc(<cnp.float32_t *> cnp.PyArray_DATA(feat),
                              <cnp.float32_t *> cnp.PyArray_DATA(query),
                              <cnp.float32_t *> cnp.PyArray_DATA(featl2norm),
                              D, N,
                              <cnp.float32_t *> cnp.PyArray_DATA(dist))
    return dist



#!/usr/bin/env python
# coding: utf-8

"""
   File Name: sh.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Mon Jul 27 10:22:06 2015 CST
"""
DESCRIPTION = """
"""

import logging

import numpy as np

from hdidx.indexer import Indexer
from hdidx.encoder import SHEncoder
from hdidx.util import Profiler
from hdidx.storage import createStorage

import hdidx._cext as cext

BIT_CNT_MAP = np.array([bin(i).count("1") for i in xrange(256)], np.uint16)


class SHIndexer(Indexer):
    def __init__(self):
        Indexer.__init__(self)
        self.encoder = SHEncoder()
        self.set_storage()

    def __del__(self):
        pass

    def build(self, pardic=None):
        self.encoder.build(pardic)

    def set_storage(self, storage_type='mem', storage_parm=None):
        self.storage = createStorage(storage_type, storage_parm)

    def add(self, vals, keys=None):
        num_vals = vals.shape[0]
        if keys is None:
            num_base_items = self.storage.get_num_items()
            keys = np.arange(num_base_items, num_base_items + num_vals,
                             dtype=np.int32)
        else:
            keys = np.array(keys, dtype=np.int32).reshape(-1)

        start_id = 0
        for start_id in range(0, num_vals, self.BLKSIZE):
            cur_num = min(self.BLKSIZE, num_vals - start_id)
            logging.info("%8d/%d: %d" % (start_id, num_vals, cur_num))
            codes = self.encoder.encode(vals[start_id:start_id+cur_num, :])
            self.storage.add(codes, keys[start_id:start_id+cur_num])

    def remove(self, keys):
        raise Exception(self.ERR_UNIMPL)

    @staticmethod
    def hammingDist(B1, B2):
        """
        Compute hamming distance between two sets of samples (B1, B2)

        Dh=hammingDist(B1, B2);

        Input
        B1, B2: compact bit vectors. Each datapoint is one row.
        size(B1) = [ndatapoints1, nwords]
        size(B2) = [ndatapoints2, nwords]
        It is faster if ndatapoints1 < ndatapoints2

        Output
        Dh = hamming distance.
        size(Dh) = [ndatapoints1, ndatapoints2]

        example query
        Dhamm = hammingDist(B2, B1);
        this will give the same result than:
        Dhamm = distMat(U2>0, U1>0).^2;
        the size of the distance matrix is:
        size(Dhamm) = [Ntest x Ntraining]
        """

        if B1.ndim == 1:
            B1 = B1.reshape((1, -1))

        if B2.ndim == 1:
            B2 = B2.reshape((1, -1))

        npt1, dim1 = B1.shape
        npt2, dim2 = B2.shape

        if dim1 != dim2:
            raise Exception("Dimension not consists: %d, %d" % (dim1, dim2))

        Dh = np.zeros((npt1, npt2), np.uint16)

        for i in xrange(npt1):
            Dh[i, :] = BIT_CNT_MAP[np.bitwise_xor(B1[i, :], B2)].sum(1)

        return Dh

    @staticmethod
    def hammingDist2(B1, B2):
        """
        Compute hamming distance between two sets of samples (B1, B2)

        Dh=hammingDist(B1, B2);

        Input
        B1, B2: compact bit vectors. Each datapoint is one row.
        size(B1) = [ndatapoints1, nwords]
        size(B2) = [ndatapoints2, nwords]
        It is faster if ndatapoints1 < ndatapoints2

        Output
        Dh = hamming distance.
        size(Dh) = [ndatapoints1, ndatapoints2]

        example query
        Dhamm = hammingDist(B2, B1);
        this will give the same result than:
        Dhamm = distMat(U2>0, U1>0).^2;
        the size of the distance matrix is:
        size(Dhamm) = [Ntest x Ntraining]
        """

        if B1.ndim == 1:
            B1 = B1.reshape((1, -1))

        if B2.ndim == 1:
            B2 = B2.reshape((1, -1))

        npt1, dim1 = B1.shape
        npt2, dim2 = B2.shape

        if dim1 != dim2:
            raise Exception("Dimension not consists: %d, %d" % (dim1, dim2))

        Dh = cext.hamming(B1, B2)

        return Dh

    def search(self, queries, topk=None, **kwargs):
        nq = queries.shape[0]
        nbits = self.encoder.ecdat['nbits']

        # qry_codes = self.encoder.encode(queries)
        db_codes = self.storage.get_codes()
        idsquerybase = self.storage.get_keys()

        dis = np.ones((nq, topk), np.single) * np.inf
        ids = np.ones((nq, topk), np.int32) * -1

        profiler = Profiler()
        interval = 100 if nq >= 100 else 10
        time_total = 0.0    # total time for all queries
        logging.info('Start Querying ...')
        for qry_id in range(nq):
            profiler.start("encoding")  # time for computing the distances
            qry_code = self.encoder.encode(queries[qry_id:qry_id+1])
            profiler.end()

            profiler.start("distance")  # time for computing the distances
            disquerybase = self.hammingDist2(qry_code, db_codes).reshape(-1)
            profiler.end()

            profiler.start("knn")       # time for finding the kNN
            cur_ids = cext.knn_count(disquerybase, nbits, topk)
            profiler.end()

            profiler.start("result")    # time for getting final result
            ids[qry_id, :] = idsquerybase[cur_ids]
            dis[qry_id, :] = disquerybase[cur_ids]
            profiler.end()

            if (qry_id+1) % interval == 0:
                time_total += profiler.sum_overall()
                logging.info(
                    '\t%d/%d: %.3fms per query' %
                    (qry_id+1, nq, profiler.sum_average() * 1000))
                logging.info("\t\t%s" % profiler.str_average())
                profiler.reset()
        logging.info('Querying Finished!')
        time_total += profiler.sum_overall()
        logging.info("Average querying time: %.3fms" % (time_total * 1000 / nq))

        return ids, dis



#!/usr/bin/env python
# coding: utf-8

"""
   File Name: indexer
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Tue Nov  4 09:07:38 2014 CST
"""
DESCRIPTION = """
"""


class Indexer(object):
    BLKSIZE = 16384

    def __init__(self):
        self.ERR_INSTAN = "Instance of `Indexer` is not allowed!"
        self.ERR_UNIMPL = "Unimplemented method!"

    def __del__(self):
        pass

    def build(self, vals=None, labels=None):
        """
        Build the indexer based on given training data
        """
        raise Exception(self.ERR_INSTAN)

    def load(self, path):
        """
        Load indexer information from file
        """
        self.encoder.load(path)

    def save(self, path):
        """
        Save the information related to the indexer itself
        """
        self.encoder.save(path)

    def set_storage(self, storage_type='mem', storage_parm=None):
        """
        Set up the backend storage engine
        """
        raise Exception(self.ERR_INSTAN)

    def add(self, vals, keys):
        """
        Add one or more items to the indexer
        """
        raise Exception(self.ERR_INSTAN)

    def remove(self, keys):
        """
        Remove one or more items from the indexer
        """
        raise Exception(self.ERR_INSTAN)

    def search(self, queries, topk=None, thresh=None):
        """
        Search in the indexer for `k` nearest neighbors or
        neighbors in a distance of `thresh`
        """
        raise Exception(self.ERR_INSTAN)

from pq import PQIndexer, IVFPQIndexer
from sh import SHIndexer



#!/usr/bin/env python
# coding: utf-8

"""
   File Name: pq.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Mon Jul 27 10:22:00 2015 CST
"""
DESCRIPTION = """
"""

import logging

import numpy as np

from hdidx.indexer import Indexer
from hdidx.encoder import PQEncoder, IVFPQEncoder
from hdidx.util import pq_knn, Profiler
from hdidx.distance import distFunc
# from hdidx.distance import fast_euclidean
from hdidx._cext import fast_euclidean
from hdidx.storage import createStorage

import hdidx._cext as cext


class PQIndexer(Indexer):
    def __init__(self):
        Indexer.__init__(self)
        self.encoder = PQEncoder()
        self.storage = None

    def __del__(self):
        pass

    def build(self, pardic=None):
        self.encoder.build(pardic)

    def set_storage(self, storage_type='mem', storage_parm=None):
        self.storage = createStorage(storage_type, storage_parm)

    def add(self, vals, keys=None):
        if self.storage is None:
            self.set_storage()

        num_vals = vals.shape[0]
        if keys is None:
            num_base_items = self.storage.get_num_items()
            keys = np.arange(num_base_items, num_base_items + num_vals,
                             dtype=np.int32)
        else:
            keys = np.array(keys, dtype=np.int32).reshape(-1)

        start_id = 0
        for start_id in range(0, num_vals, self.BLKSIZE):
            cur_num = min(self.BLKSIZE, num_vals - start_id)
            logging.info("%8d/%d: %d" % (start_id, num_vals, cur_num))

            codes = self.encoder.encode(vals[start_id:start_id+cur_num])
            self.storage.add(codes, keys[start_id:start_id+cur_num])

    def remove(self, keys):
        raise Exception(self.ERR_UNIMPL)

    def search(self, queries, topk=None,  **kwargs):
        nq = queries.shape[0]

        dsub = self.encoder.ecdat['dsub']
        nsubq = self.encoder.ecdat['nsubq']
        ksub = self.encoder.ecdat['ksub']
        centroids = self.encoder.ecdat['centroids']

        distab = np.zeros((nsubq, ksub), np.single)
        dis = np.ones((nq, topk), np.single) * np.inf
        ids = np.ones((nq, topk), np.int32) * -1

        profiler = Profiler()
        interval = 100 if nq >= 100 else 10
        time_total = 0.0    # total time for all queries
        logging.info('Start Querying ...')
        for qry_id in xrange(nq):
            profiler.start("distab")    # time for computing distance table
            # pre-compute the table of squared distance to centroids
            for qnt_id in range(nsubq):
                vsub = queries[qry_id:qry_id+1, qnt_id*dsub:(qnt_id+1)*dsub]
                distab[qnt_id:qnt_id+1, :] = distFunc['euclidean'](
                    centroids[qnt_id], vsub)
            profiler.end()

            profiler.start("distance")  # time for computing the distances
            # add the tabulated distances to construct the distance estimators
            idsquerybase, disquerybase = self.sumidxtab(distab)
            profiler.end()

            profiler.start("knn")       # time for finding the kNN
            realk = min(disquerybase.shape[0], topk)
            cur_ids = pq_knn(disquerybase, realk)
            profiler.end()

            profiler.start("result")    # time for getting final result
            ids[qry_id, :] = idsquerybase[cur_ids]
            dis[qry_id, :] = disquerybase[cur_ids]
            profiler.end()

            if (qry_id+1) % interval == 0:
                time_total += profiler.sum_overall()
                logging.info(
                    '\t%d/%d: %.3fms per query' %
                    (qry_id+1, nq, profiler.sum_average() * 1000))
                logging.info("\t\t%s" % profiler.str_average())
                profiler.reset()
        logging.info('Querying Finished!')
        time_total += profiler.sum_overall()
        logging.info("Average querying time: %.3fms" % (time_total * 1000 / nq))

        return ids, dis

    def sumidxtab(self, D):
        """
        Compute distance to database items based on distances to centroids.
            D: nsubq x ksub
        """

        ids = self.storage.get_keys()
        dis = cext.sumidxtab_core(D, self.storage.get_codes())

        return np.array(ids), np.array(dis)

        """
        Deprecated code
        """
        # num_base_items = self.storage.get_num_items()

        # dis = np.zeros(num_base_items)
        # ids = np.arange(0)

        # start_id = 0
        # for keys, blk in self.storage:
        #     cur_num = blk.shape[0]
        #     # dis[start_id:start_id+cur_num] = self.sumidxtab_core(D, blk)
        #     dis[start_id:start_id+cur_num] = cext.sumidxtab_core(D, blk)
        #     start_id += cur_num
        #     ids = np.hstack((ids, keys))

        # return ids, dis

    @classmethod
    def sumidxtab_core(cls, D, blk):
        # return 0
        return [sum([D[j, blk[i, j]] for j in range(D.shape[0])])
                for i in range(blk.shape[0])]


class IVFPQIndexer(PQIndexer):
    def __init__(self):
        PQIndexer.__init__(self)
        self.encoder = IVFPQEncoder()
        self.storage = None

    def __del__(self):
        pass

    def build(self, pardic=None):
        self.encoder.build(pardic)

    def set_storage(self, storage_type='mem', storage_parm=None):
        storage_parm['coarsek'] = self.encoder.ecdat['coarsek']
        self.storage = createStorage(storage_type, storage_parm)

    def add(self, vals, keys=None):
        if self.storage is None:
            self.set_storage()

        num_vals = vals.shape[0]
        if keys is None:
            num_base_items = sum([ivf.get_num_items() for ivf in self.storage])
            keys = np.arange(num_base_items, num_base_items + num_vals,
                             dtype=np.int32)
        else:
            keys = np.array(keys, dtype=np.int32).reshape(-1)

        start_id = 0
        logging.info("Building indexes - BEGIN")
        for start_id in range(0, num_vals, self.BLKSIZE):
            cur_num = min(self.BLKSIZE, num_vals - start_id)
            end_id = start_id + cur_num
            logging.info("%8d/%d: %d" % (start_id, num_vals, cur_num))

            cids, codes = self.encoder.encode(vals[start_id:end_id, :])

            coarsek = self.encoder.ecdat['coarsek']
            for ivfidx in xrange(coarsek):
                self.storage[ivfidx].add(
                    codes[cids == ivfidx, :],
                    keys[start_id:end_id][cids == ivfidx])
        logging.info("Building indexes - DONE!")

    def remove(self, keys):
        raise Exception(self.ERR_UNIMPL)

    def search(self, queries, topk=None, **kwargs):
        nn_coa = kwargs.get('nn_coa', 8)
        nq = queries.shape[0]

        dsub = self.encoder.ecdat['dsub']
        nsubq = self.encoder.ecdat['nsubq']
        ksub = self.encoder.ecdat['ksub']
        centroids = self.encoder.ecdat['centroids']
        coa_centroids = self.encoder.ecdat['coa_centroids']

        centroids_l2norm = []
        for i in xrange(nsubq):
            centroids_l2norm.append((centroids[i] ** 2).sum(1))
        coa_centroids_l2norm = (coa_centroids ** 2).sum(1)

        distab = np.zeros((nsubq, ksub), np.single)
        dis = np.ones((nq, topk), np.single) * np.inf
        ids = np.ones((nq, topk), np.int32) * -1

        profiler = Profiler()
        interval = 100 if nq >= 100 else 10
        time_total = 0.0    # total time for all queries
        logging.info('Start Querying ...')
        for qry_id in xrange(nq):
            # Here `copy()` can ensure that you DONOT modify the queries
            query = queries[qry_id:qry_id+1, :].copy()
            profiler.start("coa_knn")
            coa_dist = fast_euclidean(
                coa_centroids, query, coa_centroids_l2norm).reshape(-1)
            # profiler.end()
            # profiler.start("coa_knn_knn")
            coa_knn = pq_knn(coa_dist, nn_coa)
            profiler.end()

            profiler.start("distab+distance")
            query = query - coa_centroids[coa_knn, :]
            v_idsquerybase = []
            v_disquerybase = []
            for coa_idx in range(nn_coa):
                # pre-compute the table of squared distance to centroids
                for qnt_id in range(nsubq):
                    vsub = query[coa_idx:coa_idx+1,
                                 qnt_id*dsub:(qnt_id+1)*dsub]
                    distab[qnt_id:qnt_id+1, :] = fast_euclidean(
                        centroids[qnt_id], vsub, centroids_l2norm[qnt_id])

                # construct the distance estimators from tabulated distances
                idsquerybase, disquerybase = self.sumidxtab(
                    distab, coa_knn[coa_idx])
                v_idsquerybase.append(idsquerybase)
                v_disquerybase.append(disquerybase)

            idsquerybase = np.hstack(tuple(v_idsquerybase))
            disquerybase = np.hstack(tuple(v_disquerybase))
            profiler.end()

            profiler.start("knn")       # time for finding the kNN
            realk = min(disquerybase.shape[0], topk)
            cur_ids = pq_knn(disquerybase, realk)
            profiler.end()

            profiler.start("result")    # time for getting final result
            ids[qry_id, :realk] = idsquerybase[cur_ids]
            dis[qry_id, :realk] = disquerybase[cur_ids]
            profiler.end()

            if (qry_id+1) % interval == 0:
                time_total += profiler.sum_overall()
                logging.info(
                    '\t%d/%d: %.3fms per query' %
                    (qry_id+1, nq, profiler.sum_average() * 1000))
                logging.info("\t\t%s" % profiler.str_average())
                profiler.reset()
        logging.info('Querying Finished!')
        time_total += profiler.sum_overall()
        logging.info("Average querying time: %.3fms" % (time_total * 1000 / nq))

        return ids, dis

    def sumidxtab(self, D, ivfidx):
        """
        Compute distance to database items based on distances to centroids.
            D: nsubq x ksub
        """

        ids = self.storage[ivfidx].get_keys()
        if ids.shape[0] == 0:
            dis = np.ndarray(0)
        else:
            dis = cext.sumidxtab_core(D, self.storage[ivfidx].get_codes())

        return np.array(ids), np.array(dis)

        """
        Deprecated code
        """
        num_candidates = self.storage[ivfidx].get_num_items()
        dis = np.zeros(num_candidates)
        ids = np.arange(0)

        start_id = 0

        for keys, blk in self.storage[ivfidx]:
            cur_num = blk.shape[0]
            # dis[start_id:start_id+cur_num] = self.sumidxtab_core(D, blk)
            dis[start_id:start_id+cur_num] = cext.sumidxtab_core(D, blk)
            start_id += cur_num
            ids = np.hstack((ids, keys))

        return ids, dis



#!/usr/bin/env python
# coding: utf-8

"""
   File Name: sh.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Fri Jul 31 20:26:32 2015 CST
"""
DESCRIPTION = """
"""

import logging

import numpy as np

from hdidx.encoder import Encoder
from hdidx.util import eigs


class SHEncoder(Encoder):
    def __init__(self):
        Encoder.__init__(self)

    def __del__(self):
        pass

    def build(self, pardic=None):
        # training data
        X = pardic['vals']
        # the number of subquantizers
        nbits = pardic['nbits']
        # the number of items in one block
        blksize = pardic.get('blksize', 16384)

        [Nsamples, Ndim] = X.shape

        # algo:
        # 1) PCA
        npca = min(nbits, Ndim)
        pc, l = eigs(np.cov(X.T), npca)
        X = X.dot(pc)   # no need to remove the mean

        # 2) fit uniform distribution
        eps = np.finfo(float).eps
        mn = np.percentile(X, 5)
        mx = np.percentile(X, 95)
        mn = X.min(0) - eps
        mx = X.max(0) + eps

        # 3) enumerate eigenfunctions
        R = mx - mn
        maxMode = np.ceil((nbits+1) * R / R.max())
        nModes = maxMode.sum() - maxMode.size + 1
        modes = np.ones((nModes, npca))
        m = 0
        for i in xrange(npca):
            modes[m+1:m+maxMode[i], i] = np.arange(1, maxMode[i]) + 1
            m = m + maxMode[i] - 1
        modes = modes - 1
        omega0 = np.pi / R
        omegas = modes * omega0.reshape(1, -1).repeat(nModes, 0)
        eigVal = -(omegas ** 2).sum(1)
        ii = (-eigVal).argsort()
        modes = modes[ii[1:nbits+1], :]

        """
        Initializing encoder data
        """
        ecdat = dict()
        ecdat['nbits'] = nbits
        ecdat['pc'] = pc
        ecdat['mn'] = mn
        ecdat['mx'] = mx
        ecdat['modes'] = modes
        ecdat['blksize'] = blksize
        self.ecdat = ecdat

    @staticmethod
    def compactbit(b):
        nSamples, nbits = b.shape
        nwords = (nbits + 7) / 8
        B = np.hstack([np.packbits(b[:, i*8:(i+1)*8][:, ::-1], 1)
                       for i in xrange(nwords)])
        residue = nbits % 8
        if residue != 0:
            B[:, -1] = np.right_shift(B[:, -1], 8 - residue)

        return B

    def encode(self, vals):
        X = vals
        if X.ndim == 1:
            X = X.reshape((1, -1))

        Nsamples, Ndim = X.shape
        nbits = self.ecdat['nbits']
        mn = self.ecdat['mn']
        mx = self.ecdat['mx']
        pc = self.ecdat['pc']
        modes = self.ecdat['modes']

        X = X.dot(pc)
        X = X - mn.reshape((1, -1))
        omega0 = 0.5 / (mx - mn)
        omegas = modes * omega0.reshape((1, -1))

        U = np.zeros((Nsamples, nbits))
        for i in range(nbits):
            omegai = omegas[i, :]
            ys = X * omegai + 0.25
            ys -= np.floor(ys)
            yi = np.sum(ys < 0.5, 1)
            U[:, i] = yi

        b = np.require(U % 2 == 0, dtype=np.int)
        B = self.compactbit(b)
        return B



#!/usr/bin/env python
# coding: utf-8

"""
   File Name: encoder
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Fri Jul 31 20:18:30 2015 CST
"""
DESCRIPTION = """
"""

import cPickle as pickle


class Encoder(object):
    """
    Encoder maps original data to hash codes.
    """
    def __init__(self):
        self.ERR_INSTAN = "Instance of `Encoder` is not allowed!"
        self.ERR_UNIMPL = "Unimplemented method!"
        pass

    def __del__(self):
        pass

    def build(self, vals=None, labels=None):
        """
        Build the encoder based on given training data.
        """
        raise Exception(self.ERR_INSTAN)

    def load(self, path):
        """
        Load encoder information from file.
        """
        with open(path, 'rb') as pklf:
            self.ecdat = pickle.load(pklf)

    def save(self, path):
        """
        Save the encoder information.
        """
        with open(path, 'wb') as pklf:
            pickle.dump(self.ecdat, pklf, protocol=2)

    def encode(self, vals):
        """
        Map `vals` to hash codes.
        """
        raise Exception(self.ERR_INSTAN)

from pq import PQEncoder, IVFPQEncoder
from sh import SHEncoder



#!/usr/bin/env python
# coding: utf-8

"""
   File Name: lmdb_storage.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Sat Aug  8 16:11:07 2015 CST
"""
DESCRIPTION = """
"""

import cPickle as pickle
from struct import pack, unpack
import numpy as np

from hdidx.storage import MemStorage


class LMDBAccessor(object):
    def __init__(self, env, dbname):
        self.env = env
        self.db = env.open_db(dbname)

    def close(self):
        return
        self.db.close()

    def setkv(self, key, val, ktype, vtype):
        with self.env.begin(write=True) as txt:
            txt.cursor(self.db).put(pack(ktype, key), pack(vtype, key))

    def setk(self, key, val, ktype):
        with self.env.begin(write=True) as txt:
            txt.cursor(self.db).put(pack(ktype, key), val)

    def setv(self, key, val, vtype):
        with self.env.begin(write=True) as txt:
            txt.cursor(self.db).put(key, pack(vtype, val))

    def set(self, key, val):
        with self.env.begin(write=True) as txt:
            txt.cursor(self.db).put(key, val)

    def getkv(self, key, ktype, vtype):
        with self.env.begin(write=True) as txt:
            val = txt.cursor(self.db).get(pack(ktype, key))
            if val is None:
                return None
            return unpack(vtype, val)[0]

    def getk(self, key, ktype):
        with self.env.begin(write=True) as txt:
            return txt.cursor(self.db).get(pack(ktype, key))

    def getv(self, key, vtype):
        with self.env.begin(write=True) as txt:
            val = txt.cursor(self.db).get(key)
            if val is None:
                return None
            return unpack(vtype, val)[0]

    def get(self, key):
        with self.env.begin(write=True) as txt:
            return txt.cursor(self.db).get(key)

    def setkivi(self, key, val):
        self.setkv(key, val, 'i', 'i')

    def setki(self, key, val):
        self.setk(key, val, 'i')

    def setvi(self, key, val):
        self.setv(key, val, 'i')

    def getkivi(self, key):
        return self.getkv(key, 'i', 'i')

    def getki(self, key):
        return self.getk(key, 'i')

    def getvi(self, key):
        return self.getv(key, 'i')


class LMDBStorage(MemStorage):
    def __init__(self, env, clear, ivfidx=None):
        MemStorage.__init__(self)

        self.env = env
        dbname = 'db%06d' % ivfidx if ivfidx else 'db'
        self.db = LMDBAccessor(self.env, dbname)

        if clear:
            self.clear()

        with self.env.begin() as txn:
            self.num_items = txn.stat(self.db.db)['entries']
            if self.num_items > 0:
                cursor = txn.cursor(self.db.db)
                keys = []
                codes = []
                for key, code in cursor:
                    keys.append(unpack('i', key)[0])
                    codes.append(pickle.loads(code))
                self.keys = np.array(keys)
                self.codes = np.vstack(tuple(codes))

    def __del__(self):
        self.db.close()

    def add(self, codes, keys):
        num_new_items = MemStorage.add(self, codes, keys)
        for idx in xrange(num_new_items):
            self.db.setki(keys[idx], pickle.dumps(codes[idx], protocol=2))

    def clear(self):
        with self.env.begin(write=True) as txt:
            txt.drop(self.db.db, False)

"""
    def __iter__(self):
        self.iter_txt = self.env.begin()
        self.cursor_keys = self.iter_txt.cursor(self.db_keys.db)
        self.cursor_vals = self.iter_txt.cursor(self.db_vals.db)
        return self

    def next(self):
        status_keys = self.cursor_keys.next()
        status_vals = self.cursor_vals.next()
        if status_keys and status_vals:
            keys = pickle.loads(self.cursor_keys.value())
            vals = pickle.loads(self.cursor_vals.value())
            return keys, vals
        else:
            # self.iter_txt.close()
            raise StopIteration
"""



#!/usr/bin/env python
# coding: utf-8

"""
   File Name: hdidx
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Tue Nov  4 09:00:29 2014 CST
"""
DESCRIPTION = """
"""
import indexer
import encoder
import storage
import distance



#!/usr/bin/env python
# coding: utf-8

#########################################################################
#########################################################################

"""
   File Name: util.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Tue Nov  4 09:38:24 2014 CST
"""
DESCRIPTION = """
"""

import os
import logging
# distance
from distance import distFunc
import bottleneck
from scipy.io import loadmat
import numpy as np

# profiling
import time


DO_NORM = {
    "cosine": True,
    "euclidean": False,
}


class HDIdxException(Exception):
    """
    HDIdx Exception
    """


"""
Math
"""


def eigs(X, npca):
    l, pc = np.linalg.eig(X)
    idx = l.argsort()[::-1][:npca]
    return pc[:, idx], l[idx]


"""
KMeans
"""

try:
    import cv2

    def kmeans(vs, ks, niter):
        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER,
                    niter, 0.01)
        flags = cv2.KMEANS_RANDOM_CENTERS
        compactness, labels, centers = cv2.kmeans(
            vs, ks, criteria, 1, flags)
        return centers
except ImportError:
    logging.warn("Cannot find OpenCV, using `kmeans` from SciPy instead.")
    from scipy.cluster import vq

    def kmeans(vs, ks, niter):
        centers, labels = vq.kmeans2(vs, ks, niter)
        return centers


# finding nearest neighbor


def pq_kmeans_assign(centroids, query):
    dist = distFunc['euclidean'](centroids, query)
    return dist.argmin(1)


def pq_knn(dist, topk):
    ids = bottleneck.argpartsort(dist, topk)[:topk]
    ids = ids[dist[ids].argsort()]
    return ids


# Profiling
START_TIME = 0


def tic():
    global START_TIME
    START_TIME = time.time()


def toc():
    return time.time() - START_TIME


class Profiler(object):
    """ Profiling the running time of code snippet.
    """
    class Record(object):
        __slots__ = ["name", "time", "count", "t0"]

        def __init__(self, name):
            self.name = name
            self.reset()

        def reset(self):
            self.time = 0.0
            self.count = 0
            self.t0 = None

        def start(self):
            self.t0 = time.time()

        def end(self):
            self.time += (time.time() - self.t0)
            self.count += 1
            self.t0 = None

        def average(self):
            return self.time / self.count if self.count > 0 else 0

    __slots__ = ["records",
                 "cur_record",
                 "name_stack"]

    def __init__(self):
        self.reset()

    def start(self, name):
        """
        Start the timer.
        `name` is the description of the current code snippet.
        """
        if name not in self.records:
            self.records[name] = Profiler.Record(name)
        self.cur_record = self.records[name]
        self.name_stack.append(name)
        self.cur_record.start()

    def end(self, name=None):
        """
        Calculate the time costs of the current code snippet.
        """
        if name is not None and name != self.name_stack[-1]:
            raise Exception("name '%s' should be '%s'" %
                            (name, self.name_stack[-1]))
        self.cur_record.end()
        self.name_stack.pop()

    def sum_overall(self):
        """
        Return the sum of overall time costs for each code snippet.
        """
        return sum([rec.time for name, rec in self.records.iteritems()])

    def sum_average(self):
        """
        Return the sum of average time costs for each code snippet.
        """
        return sum([rec.average() for name, rec in self.records.iteritems()])

    def str_overall(self, fmt="%s: %.3fms"):
        """
        Return the overall time costs for each code snippet as string.
        """

        return ";\t".join([fmt % (name, rec.time * 1000)
                           for name, rec in self.records.iteritems()])

    def str_average(self, fmt="%s: %.3fms"):
        """
        Return the average time costs for each code snippet as string.
        """
        return ";\t".join([fmt % (name, rec.average() * 1000)
                           for name, rec in self.records.iteritems()])

    def reset(self):
        """
        Reset the time costs and counters.
        """
        self.records = {}
        self.name_stack = []
        self.cur_record = None


def normalize(feat, ln=2):
    if ln is 1:
        return feat / feat.sum(1).reshape(-1, 1)
    elif ln > 0:
        return feat / ((feat**ln).sum(1)**(1.0/ln)).reshape(-1, 1)
    else:
        raise Exception("Unsupported norm: %d" % ln)


def tokey(item):
    """
    Key function for sorting filenames
    """
    return int(item.split("_")[-1].split(".")[0])


class Reader(object):
    def __init__(self, featdir):
        self.v_fname = sorted(os.listdir(featdir), key=tokey)
        self.next_id = 0
        self.featdir = featdir

    def get_next(self):
        logging.info("Reader - load %d" % self.next_id)
        feat = loadmat(
            os.path.join(self.featdir, self.v_fname[self.next_id]))['feat']
        self.next_id += 1
        return feat



#!/usr/bin/env python
# coding: utf-8

"""
   File Name: pq.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Fri Jul 31 20:26:32 2015 CST
"""
DESCRIPTION = """
"""

import logging

import numpy as np

from hdidx.encoder import Encoder
from hdidx.util import kmeans, pq_kmeans_assign


class PQEncoder(Encoder):
    """
    Product Quantization Encoder
    """
    def __init__(self):
        super(PQEncoder, self).__init__()

    def __del__(self):
        pass

    def build(self, pardic=None):
        # training data
        vals = pardic['vals']
        # the number of subquantizers
        nsubq = pardic['nsubq']
        # the number bits of each subquantizer
        nsubqbits = pardic.get('nsubqbits', 8)
        # the number of items in one block
        blksize = pardic.get('blksize', 16384)

        # vector dimension
        dim = vals.shape[1]
        # dimension of the subvectors to quantize
        dsub = dim / nsubq
        # number of centroids per subquantizer
        ksub = 2 ** nsubqbits

        """
        Initializing indexer data
        """
        ecdat = dict()
        ecdat['nsubq'] = nsubq
        ecdat['ksub'] = ksub
        ecdat['dsub'] = dsub
        ecdat['blksize'] = blksize
        ecdat['centroids'] = [None for q in range(nsubq)]

        logging.info("Building codebooks in subspaces - BEGIN")
        for q in range(nsubq):
            logging.info("\tsubspace %d/%d" % (q, nsubq))
            vs = np.require(vals[:, q*dsub:(q+1)*dsub],
                            requirements='C', dtype=np.float32)
            ecdat['centroids'][q] = kmeans(vs, ksub, niter=100)
        logging.info("Building codebooks in subspaces - DONE")

        self.ecdat = ecdat

    def encode(self, vals):
        dsub = self.ecdat['dsub']
        nsubq = self.ecdat['nsubq']
        centroids = self.ecdat['centroids']

        num_vals = vals.shape[0]
        codes = np.zeros((num_vals, nsubq), np.uint8)
        for q in range(nsubq):
            vsub = vals[:, q*dsub:(q+1)*dsub]
            codes[:, q] = pq_kmeans_assign(centroids[q], vsub)
        return codes


class IVFPQEncoder(PQEncoder):
    """
    Inverted-File Product Quantization Encoder
    """
    def __init__(self):
        super(IVFPQEncoder, self).__init__()

    def __del__(self):
        pass

    def build(self, pardic=None):
        pardic['vals'] = pardic['vals'].copy()
        # training data
        vals = pardic['vals']
        # the number of coarse centroids
        coarsek = pardic['coarsek']

        logging.info('Building coarse quantizer - BEGIN')
        coa_centroids = kmeans(vals, coarsek, niter=100)
        cids = pq_kmeans_assign(coa_centroids, vals)
        logging.info('Building coarse quantizer - DONE')

        pardic['vals'] -= coa_centroids[cids, :]
        super(IVFPQEncoder, self).build(pardic)

        self.ecdat['coa_centroids'] = coa_centroids
        self.ecdat['coarsek'] = coarsek

    def encode(self, vals):
        # Here `copy()` can ensure that you DONOT modify the vals
        vals = vals.copy()

        coa_centroids = self.ecdat['coa_centroids']

        cids = pq_kmeans_assign(coa_centroids, vals)
        vals -= coa_centroids[cids, :]

        codes = super(IVFPQEncoder, self).encode(vals)
        return cids, codes



#!/usr/bin/env python
# coding: utf-8

#########################################################################
#########################################################################

"""
   File Name: test_cext.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Wed Nov  5 16:41:24 2014 CST
"""
DESCRIPTION = """
"""

import unittest

# import time
# import logging
import numpy as np
# import ipdb

import hdidx._cext as cext
from hdidx.indexer.sh import SHIndexer as sh


class TestCExt(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def test_sumidxtab_core(self):
        nsq = 8
        ksub = 256
        cur_num = 10

        for iround in range(10):
            raw_D = np.random.random((nsq, ksub))
            raw_blk = np.random.random_integers(0, ksub-1, (cur_num, nsq))
            D = np.require(raw_D, np.float32, "C")
            blk = np.require(raw_blk, np.uint8, "C")

            self.assertLessEqual(np.abs(raw_D - D).sum(),  1e-4)
            self.assertEqual(np.abs(raw_blk - blk).sum(),  0)

            py_res = self.sumidxtab_core(D, blk)
            c_res = cext.sumidxtab_core(D, blk)
            self.assertLessEqual(np.abs(py_res - c_res).sum(),  1e-4)

    def sumidxtab_core(cls, D, blk):
        return np.require([sum([D[j, blk[i, j]] for j in range(D.shape[0])])
                           for i in range(blk.shape[0])], np.float32)

    def test_hamming_core(self):
        Q = (np.random.random((100, 4)) * 255).astype(np.uint8)
        D = (np.random.random((1000, 4)) * 255).astype(np.uint8)
        D1 = cext.hamming(Q, D)
        D2 = sh.hammingDist(Q, D)
        D3 = sh.hammingDist2(Q, D)

        self.assertEqual(self.ndarray_diff(D1, D2), 0)
        self.assertEqual(self.ndarray_diff(D1, D3), 0)

    def hamming_core(cls, D, blk):
        return np.require([sum([D[j, blk[i, j]] for j in range(D.shape[0])])
                           for i in range(blk.shape[0])], np.float32)

    def ndarray_diff(cls, A, B):
        return np.abs(A - B).max()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# coding: utf-8

#########################################################################
#########################################################################

"""
   File Name: storage
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Tue Nov  4 11:25:03 2014 CST
"""
DESCRIPTION = """
"""

import lmdb


class Storage(object):
    def __init__(self):
        self.keys = None
        self.codes = None
        self.num_emptys = -1
        self.num_items = -1

    def get_num_items(self):
        return self.num_items

    def get_num_emptys(self):
        return self.num_emptys

    def get_keys(self):
        return self.keys

    def get_codes(self):
        return self.codes

    def __iter__(self):
        raise Exception("Instance of `Storage` is not allowed!")

    def next(self):
        raise Exception("Instance of `Storage` is not allowed!")

    def clear(self):
        raise Exception("Instance of `Storage` is not allowed!")


from mem_storage import MemStorage
from lmdb_storage import LMDBStorage


STORAGE_DIC = {
    'mem':  MemStorage,
    'lmdb': LMDBStorage,
}


def createStorage(storage_type, storage_parm=None):
    coarsek = storage_parm.get('coarsek', 0) if storage_parm else 0

    if storage_type == 'mem':
        # create inverted file storage
        if coarsek > 0:
            return [MemStorage() for i in xrange(coarsek)]
        # create normal storage
        else:
            return MemStorage()
    elif storage_type == 'lmdb':
        path = storage_parm['path']
        clear = storage_parm.get('clear', False)

        # create inverted file storage
        if coarsek > 0:
            env = lmdb.open(path, map_size=2**30, sync=False, max_dbs=coarsek)
            return [LMDBStorage(env, clear, i) for i in xrange(coarsek)]
        # create normal storage
        else:
            env = lmdb.open(path, map_size=2**30, sync=False, max_dbs=1)
            return LMDBStorage(env, clear)
    else:
        raise Exception('Wroing storage type: %s' % storage_type)



#!/usr/bin/env python
# coding: utf-8

"""
   File Name: mem_storage.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Sat Aug  8 16:11:01 2015 CST
"""
DESCRIPTION = """
"""

import itertools
import numpy as np
from hdidx.storage import Storage


class MemStorage(Storage):
    def __init__(self):
        Storage.__init__(self)
        self.keys = np.arange(0, dtype=np.int32)
        self.codes = None
        self.num_items = 0

    def add(self, codes, keys):
        num_new_items = codes.shape[0]
        self.keys = np.hstack((self.keys, keys))
        if self.codes is None:
            self.codes = codes
        else:
            self.codes = np.vstack((self.codes, codes))
        self.num_items += num_new_items
        return num_new_items

    def __iter__(self):
        return itertools.izip(self.keys, self.vals)



#!/usr/bin/env python
# coding: utf-8
"""
   File Name: setup.py
      Author: Wan Ji
      E-mail: wanji@live.com
  Created on: Tue Nov  4 08:55:15 2014 CST
 Description:
"""


try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

try:
    from pypandoc import convert

    def read_md(fpath):
        return convert(fpath, 'rst')
except ImportError:
    print "warning: pypandoc module not found, DONOT convert Markdown to RST"

    def read_md(fpath):
        with open(fpath, 'r') as fp:
            return fp.read()

import numpy
from Cython.Distutils import build_ext

setup(name='hdidx',
      version='0.2',
      author='WAN Ji',
      author_email='wanji@live.com',
      package_dir={'hdidx': 'hdidx'},
      packages=[
          'hdidx',
          'hdidx.indexer',
          'hdidx.encoder',
          'hdidx.storage',
      ],

      cmdclass={'build_ext': build_ext},
      ext_modules=[Extension(
          "hdidx._cext",
          sources=[
              "cext/_cext.pyx",
              "cext/cext.c",
          ],
          include_dirs=[numpy.get_include(), "cext"])],

      scripts=[
          'tests/test_indexer.py',
          'tests/test_cext.py',
          'tools/hdidx_build',
          'tools/hdidx_index',
      ],
      url='https://github.com/wanji/hdidx',
      license='LICENSE.md',
      description='ANN Search in High-Dimensional Spaces',

      long_description=read_md("DESC.md"),
      install_requires=[
          "numpy      >= 1.6.0",
          "scipy      >= 0.9.0",
          "bottleneck >= 0.8.0",
          "lmdb       >= 0.83",
      ],
      )
